/*
 * IDLflex -- A flexible, generic code generator for CORBA IDL
 *
 * Copyright (C) 2000  Hans Reiser, AspectIX Research Team
 *                     University of Erlangen-Nuernberg, Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License.
 * The full text of the license can be found in the LICENSE.txt file.
 *
 * $Id: InterfaceObj.java,v 1.5 2001/09/07 09:20:44 reiser Exp $
 */

package org.aspectix.IDLflex.IRObj;

import org.omg.CORBA.*;
import java.util.Vector;
import org.aspectix.IDLflex.IDLflexException;

/* Implementation class for an IDL Interface.
 *<ul>
 * <li><em>getContentList("BASE")</em>
 *     returns all base classes from which this class inherits directly
 * <li><em>getContentList("ALLBASE")</em>
 *     returns all (direct and indirect) base classes
 * <li><em>getContentList("MEMBER")</em>
 *     returns all elements defined explicitly inside this interface
 * <li><em>getContentList("ALLMEMBER")</em>
 *     returns all member elements, including the inherited ones
 *</ul>
 */
public class InterfaceObj extends IDLObject {
    Vector allBaseInterfaces;
    Vector baseInterfaces;
    Vector members;
    Vector allMembers;  // include inherited objects;
    boolean abstractif;

    public InterfaceObj(IRObject iro) {
	super(iro);  // set object reference and name
    }
    private void add_inherited_bases(Vector allbases, IDLObject obj) 
	throws IDLflexException
    {
	if( allbases.contains(obj) ) return;
	allbases.addElement( obj );
	InterfaceDef id = InterfaceDefHelper.narrow(obj.referenceObject);
	InterfaceDef[] base = id.base_interfaces();
	for(int i=0; i<base.length; i++) {
	    add_inherited_bases( allbases, ObjectCreator.get(base[i]) );
	}
    }

    public void init_members() throws IDLflexException {
	super.init_members();
	if(referenceObject==null) return; 

	InterfaceDef id = InterfaceDefHelper.narrow(referenceObject);

	InterfaceDef[] base = id.base_interfaces();
	baseInterfaces = new Vector();
	for(int i=0; i<base.length; i++) {
	    InterfaceObj iobj = (InterfaceObj)ObjectCreator.get(base[i]);
	    baseInterfaces.addElement(iobj);
	}

	allBaseInterfaces = new Vector();
	for(int i=0; i<baseInterfaces.size(); i++) {
	    add_inherited_bases( allBaseInterfaces, 
				 (IDLObject)(baseInterfaces.elementAt(i)) );
	}
	
	Contained[] cseq = id.contents( DefinitionKind.dk_all, false );
	members = new Vector();
	for(int i=0; i<cseq.length; i++) {
	    IDLObject iobj = ObjectCreator.get( cseq[i] );
	    members.addElement(iobj);
	}

	cseq = id.contents( DefinitionKind.dk_all, true );
	allMembers = new Vector();
	for(int i=0; i<cseq.length; i++) {
	    IDLObject iobj = ObjectCreator.get( cseq[i] );
	    allMembers.addElement(iobj);
	}
	
	TypeCode tc = id.describe_interface().type;
	if(tc.kind()==TCKind.tk_objref) abstractif=false;
	else if(tc.kind()==TCKind.tk_abstract_interface) abstractif=true;
	else System.err.println("InterfaceDef is neither tk_objref nor tk_abstract_interface");

    }
    public boolean getAttribute( String spec ) throws IDLflexException {
	if(spec.equalsIgnoreCase("abstractif")) return abstractif;
	return super.getAttribute( spec );
    }
    public IDLObject getContent(String spec) throws IDLflexException {
	return super.getContent(spec);
    }

    public IDLObject[] getContentList(String spec) throws IDLflexException {
	if(spec.equalsIgnoreCase("BASE")) 
	    return VectorToIDLObjects(baseInterfaces);
	else if(spec.equalsIgnoreCase("ALLBASE")) 
	    return VectorToIDLObjects(allBaseInterfaces);
	else if(spec.equalsIgnoreCase("MEMBER")) 
	    return VectorToIDLObjects(members);
	else if(spec.equalsIgnoreCase("ALLMEMBER")) 
	    return VectorToIDLObjects(allMembers);
	else 
	    return super.getContentList(spec);
    }

    // generic setter methods
    public void setAttribute(String spec, boolean attribute) 
	throws IDLflexException
    {
        if (spec.equalsIgnoreCase("abstractif")) abstractif = attribute;
        else super.getAttribute(spec);
    }
    

    public void addToContentList(String spec, IDLObject content)
	throws IDLflexException 
    {
        if (spec.equalsIgnoreCase("BASE")) {
            baseInterfaces.addElement(content);
            IDLObject[] baseList = content.getContentList("allbase");
            for (int i = 0; i < baseList.length; i++) {
                allBaseInterfaces.addElement(baseList[i]);
            }
        }
        else if (spec.equalsIgnoreCase("ALLBASE")) 
	    allBaseInterfaces.addElement(content);
        else if (spec.equalsIgnoreCase("MEMBER")) {
	    /* TODO: macht das Sinn? */
            members.addElement(content);
            if (content.is_a("InterfaceObj")) {
                IDLObject[] memberList = content.getContentList("allmember");
                for (int i = 0; i < memberList.length; i++) {
                    allMembers.addElement(memberList[i]);
                }
            }
        }
        else if (spec.equalsIgnoreCase("ALLMEMBER")) 
	    allMembers.addElement(content);
        else super.addToContentList(spec, content);
    }    


    /**
     */
    public boolean is_a(String spec, boolean inherit) {
	spec = spec.toLowerCase();
	if(spec.equalsIgnoreCase("interfaceobj")) return true;
	if(inherit) return super.is_a(spec, inherit);
	else return false;
    }
}

